/**************************************************************************
*  Copyright (c) 2015-2016 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  20.09.2015  mifi  First Version, tested with a EIR board.
*  07.11.2015  mifi  Corrected project for CrossWorks for ARM v2.3.5.
*  06.02.2016  mifi  Tested with TinyCTS v0.10.
*  12.03.2016  mifi  Tested with TinyCTS v0.11.
**************************************************************************/
#define __MAIN_C__

/*=======================================================================*/
/*  Includes                                                             */
/*=======================================================================*/
#include <string.h>

#include "tal.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

#define LED_LD1_ON()    { AT91C_BASE_PIOB->PIO_CODR = (1<<28); }
#define LED_LD1_OFF()   { AT91C_BASE_PIOB->PIO_SODR = (1<<28); } 

#define LED_LD2_ON()    { AT91C_BASE_PIOB->PIO_CODR = (1<<29); }
#define LED_LD2_OFF()   { AT91C_BASE_PIOB->PIO_SODR = (1<<29); } 

typedef void (*IRQ_FUNC)(void);

#define SYSIRQ_PRIO  7

/*=======================================================================*/
/*  Definition of all global Data                                        */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/
/* 
 * Some TASK variables like stack and task control block.
 */
static OS_STACK (LEDStack,   TASK_LED_STK_SIZE);
static OS_STACK (StartStack, TASK_START_STK_SIZE);

static os_tcb_t TCBLed;
static os_tcb_t TCBStartTask;

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  SpuriousHandler                                                      */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void SpuriousHandler (void)
{
} /* SpuriousHandler */

/*************************************************************************/
/*  SYSIrqHandler                                                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void SYSIrqHandler (void)
{
   TAL_CPU_IRQ_ENTER();

   /* Check Periodic Interval Timer Controller */
   if (AT91C_BASE_PITC->PITC_PISR & AT91C_PITC_PITS)
   {
      (void)AT91C_BASE_PITC->PITC_PIVR;
      OSTimerCallback();
   }
   
   TAL_CPU_IRQ_EXIT();
} /* SYSIrqHandler */

/*************************************************************************/
/*  InitSysTick                                                          */
/*                                                                       */
/*  Init the SysTick timer.                                              */
/*                                                                       */                                                                       
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void InitSysTick (void)
{
   static uint8_t InitDone = 0;
   uint32_t       HirResPeriod;
   
   if (0 == InitDone)
   {
      InitDone = 1;
      
      /* Disable all interrupts first */
      AT91C_BASE_AIC->AIC_IDCR = 0xFFFFFFFF;
   
      /* Set end of interrupt command */
      AT91C_BASE_AIC->AIC_EOICR = 0;

      /* Set spurious interrupt handler */ 
      AT91C_BASE_AIC->AIC_SPU = (unsigned long)SpuriousHandler;
   
      /*
       * The CPU clock is set to 48MHz by the startup code
       * from CrossWorks for ARM. Here we must only init the
       * timer. Setup the Periodic Interval Timer with a period
       * of OS_TICKS_PER_SECOND.
       */
      HirResPeriod = ((48054857 / 16) / OS_TICKS_PER_SECOND) - 1;
       
      AT91C_BASE_PITC->PITC_PIMR = (1 << 24) |    /* PIT Enable */
                                   (1 << 25) |    /* PIT Interrupt Enable */
                                   HirResPeriod;
               
      /*
       * Install the SystemIRQHandler, which will be used
       * for the timer interrupt.
       *
       * Setup System Interrupt Mode and Vector.
       * PIT's int is System Interrupt, Peripheral ID for PIT is 1
       */
      AT91C_BASE_AIC->AIC_SMR[1] = (unsigned long)SYSIRQ_PRIO;
      AT91C_BASE_AIC->AIC_SVR[1] = (unsigned long)SYSIrqHandler;
   }      
} /* InitSysTick */

/***************************************************************************/
/*  LEDInit                                                                */
/*                                                                         */
/*  In    : none                                                           */
/*  Out   : none                                                           */
/*  Return: none                                                           */
/***************************************************************************/
static void LEDInit (void)
{
   /* Switch off all LEDs PA28 = green, PA29 = yellow */
   AT91C_BASE_PIOB->PIO_SODR = (1<<28) | (1<<29);
   
   /* 
    * Set PB28 to output 
    */ 
   
   /* Enable PIO-B clock */
   AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOB);
   
   /* Enables the PIO to control the corresponding pin */      
   AT91C_BASE_PIOB->PIO_PER = (1<<28) | (1<<29);  
   
   /* Disables the pull up resistor on the I/O line */ 
   AT91C_BASE_PIOB->PIO_PPUDR = (1<<28) | (1<<29);
   
   /* Enables the output on the I/O line */   
   AT91C_BASE_PIOB->PIO_OER = (1<<28) | (1<<29); 
   
   /* Disables the Input Change Interrupt on the I/O line */
   AT91C_BASE_PIOB->PIO_IDR = (1<<28) | (1<<29); 

} /* LEDInit */

/***************************************************************************/
/*  HardwareInit                                                           */
/*                                                                         */
/*  In    : none                                                           */
/*  Out   : none                                                           */
/*  Return: none                                                           */
/***************************************************************************/
static void HardwareInit (void)
{
   /* Init SysTick */
   InitSysTick();
   
   /* Enable System Peripheral INT */
   AT91C_BASE_AIC->AIC_IECR = (1<<1);     
   
   /* Init LED */
   LEDInit();
} /* HardwareInit */

/*************************************************************************/
/*  LEDTask                                                              */
/*                                                                       */
/*  In    : task parameter                                               */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
static void LEDTask (void *p)
{
   (void)p;
   
   while(1)
   {
      LED_LD1_ON();
      OSTimeDly(TASK_LED_DELAY_MS);
      LED_LD1_OFF();
      OSTimeDly(TASK_LED_DELAY_MS);
   }
} /* LEDTask */

/*************************************************************************/
/*  StartTask                                                            */
/*                                                                       */
/*  In    : task parameter                                               */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
static void StartTask (void *p)
{
   (void)p;
   
   /*
    * The StartTask will be used to start all other tasks in the system.
    * At the end the priority will be set to "IDLE" priority.
    */
    
   /*
    * Start the System ticker
    */
   OSSysTickStart();    

   /* Create the LED task */
   OSTaskCreate(&TCBLed, LEDTask, NULL, TASK_LED_PRIORITY,
                LEDStack, sizeof(LEDStack), 
                "LEDTask");
                
   OSTaskChangePriority(TASK_START_PRIORITY_IDLE);

   while (1)
   {
      LED_LD2_ON();
      OSTimeDly(TASK_START_DELAY_MS);
      LED_LD2_OFF();
      OSTimeDly(TASK_START_DELAY_MS);
   }
} /* StartTask */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  main                                                                 */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
int main (void)
{
   /*
    * Before the application which is using the TinyCTS can be started, 
    * the hardware of the target must be initialized. The CPU is still
    * initialized by the CrossWorks for ARM startup. But the timer tick
    * must be init too. Even the GPIO can be init here.
    */
   HardwareInit();
   
   /* Init the TinyCTS */   
   OSInit();
   
   /* 
    * Create the StartTask.
    * The StartTask is the one and only task which 
    * will start all other tasks of the system.
    */
   OSTaskCreate(&TCBStartTask, StartTask, NULL, TASK_START_PRIORITY,
                StartStack, sizeof(StartStack), 
                "StartTask");
   
   /* 
    * OSStart must be the last function here.
    *  
    * Fasten your seatbelt, engine will be started...
    */        
   OSStart();
   
   /*
    * This return here make no sense.
    * But to prevent the compiler warning:
    *    "return type of 'main' is not 'int'
    * We use an int as return :-)
    */ 
   return(0); /*lint !e527*/
} /* main */

/*************************************************************************/
/*  irq_handler                                               (CTSonly)  */
/*                                                                       */
/*  This is the ARM7 irq_handler.                                        */
/*                                                                       */                                                                       
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void irq_handler (void) __attribute__((__naked__));
void irq_handler (void)
{
   /* Interrupt prologue */
   __asm__ volatile ( "sub    lr, lr,#4\n"         /* Adjust LR */                    
                      "stmfd  sp!,{r0-r12,lr}\n"   /* Save registers on IRQ stack. */ 
                      "mrs    r1, spsr\n"          /* Save SPSR */                    
                      "stmfd  sp!,{r1}" );                                               
   /********************************************/                      
                         
   IRQ_FUNC IrqFunc;
   
   /* Get the interrupt source */   
   IrqFunc = (IRQ_FUNC)AT91C_BASE_AIC->AIC_IVR;   
   if (IrqFunc != (IRQ_FUNC)0)
   {
      IrqFunc();
   }

   /* Interrupt epilogue */
   __asm__ volatile ( "ldmfd  sp!, {r1}\n"            /* Restore SPSR */
                      "msr    spsr_c, r1\n"
                      "ldr    r0, =0xFFFFF000\n"      /* End of interrupt. */
                      "str    r0, [r0, #0x130]\n"
                      "ldmfd  sp!, {r0-r12, pc}^" );  /* Restore registers and return. */ /*lint !e539*/
} /* irq_handler */

/*** EOF ***/
